@@ -11,10 +11,9 @@ try:
 except (ImportError, ValueError):
     GLIB_AVAILABLE = False
 
+import openpaperwork_core.deps
 import openpaperwork_core.fs
 
-from .. import deps
-
 
 LOGGER = logging.getLogger(__name__)
 
@@ -87,7 +86,7 @@ class _GioFileAdapter(io.RawIOBase):
             size = self.size
             if size <= 0:
                 return b""
-        assert(size > 0)
+        assert size > 0
         return self.gin.read_bytes(size).get_data()
 
     def readall(self):
@@ -127,7 +126,11 @@ class _GioFileAdapter(io.RawIOBase):
         self.gfd.truncate(size)
 
     def fileno(self):
-        raise io.UnsupportedOperation("fileno() called on Gio.File object")
+        if self.gout is not None and hasattr(self.gout, 'get_fd'):
+            return self.gout.get_fd()
+        if self.gin is not None and hasattr(self.gin, 'get_fd'):
+            return self.gin.get_fd()
+        raise io.UnsupportedOperation("fileno() on Gio.File unsupported")
 
     def isatty(self):
         return False
@@ -254,7 +257,9 @@ class Plugin(openpaperwork_core.fs.CommonFsPluginBase):
 
     def __init__(self):
         super().__init__()
-        self.vfs = Gio.Vfs.get_default()
+        self.vfs = None
+        if GLIB_AVAILABLE:
+            self.vfs = Gio.Vfs.get_default()
         self.tmp_files = set()
 
     def get_interfaces(self):
@@ -262,15 +267,13 @@ class Plugin(openpaperwork_core.fs.CommonFsPluginBase):
 
     def chkdeps(self, out: dict):
         if not GLIB_AVAILABLE:
-            out['glib'].update(deps.GLIB)
+            out['glib'].update(openpaperwork_core.deps.GLIB)
 
-    @staticmethod
-    def _is_file_uri(uri):
-        return uri.startswith("file://")
-
-    def fs_open(self, uri, mode='r'):
-        if not self._is_file_uri(uri):
-            return None
+    def fs_open(self, uri, mode='r', needs_fileno=False, **kwargs):
+        if needs_fileno:
+            # On Windows, `Gio.[Unix]OutputStream.get_fd()` doesn't seem
+            # to be available
+            return
 
         f = self.vfs.get_file_for_uri(uri)
         if ('w' not in mode and 'a' not in mode):
@@ -288,7 +291,7 @@ class Plugin(openpaperwork_core.fs.CommonFsPluginBase):
             ))
 
     def fs_exists(self, url):
-        if not self._is_file_uri(url):
+        if not GLIB_AVAILABLE:
             return None
 
         try:
@@ -304,7 +307,7 @@ class Plugin(openpaperwork_core.fs.CommonFsPluginBase):
             raise IOError(str(exc))
 
     def fs_listdir(self, url):
-        if not self._is_file_uri(url):
+        if not GLIB_AVAILABLE:
             return None
 
         try:
@@ -323,26 +326,19 @@ class Plugin(openpaperwork_core.fs.CommonFsPluginBase):
             raise IOError(str(exc))
 
     def fs_rename(self, old_url, new_url):
-        if not self._is_file_uri(old_url):
-            return None
-        if not self._is_file_uri(new_url):
-            return None
-
         try:
             old = self.vfs.get_file_for_uri(old_url)
             new = self.vfs.get_file_for_uri(new_url)
-            assert(not old.equal(new))
+            assert not old.equal(new)
             if not old.query_exists():
                 return None
             old.move(new, Gio.FileCopyFlags.NONE)
+            return True
         except GLib.GError as exc:
             LOGGER.warning("Gio.Gerror", exc_info=exc)
             raise IOError(str(exc))
 
     def fs_unlink(self, url, trash=True, **kwargs):
-        if not self._is_file_uri(url):
-            return None
-
         try:
             f = self.vfs.get_file_for_uri(url)
             if not f.query_exists():
@@ -354,7 +350,7 @@ class Plugin(openpaperwork_core.fs.CommonFsPluginBase):
                 deleted = f.delete()
                 if not deleted:
                     raise IOError("Failed to delete %s" % url)
-                return
+                return None
 
             deleted = False
             try:
@@ -384,14 +380,12 @@ class Plugin(openpaperwork_core.fs.CommonFsPluginBase):
                     LOGGER.warning("Failed to deleted %s", url, exc_info=exc)
             if not deleted:
                 raise IOError("Failed to delete %s" % url)
+            return True
         except GLib.GError as exc:
             LOGGER.warning("Gio.Gerror", exc_info=exc)
             raise IOError(str(exc))
 
     def fs_rm_rf(self, url, trash=True, **kwargs):
-        if not self._is_file_uri(url):
-            return None
-
         if self.fs_exists(url) is None:
             return None
 
@@ -427,13 +421,18 @@ class Plugin(openpaperwork_core.fs.CommonFsPluginBase):
             if not deleted:
                 self._rm_rf(f)
             LOGGER.info("%s deleted", url)
+            return True
         except GLib.GError as exc:
             LOGGER.warning("Gio.Gerror", exc_info=exc)
             raise IOError(str(exc))
 
     def _rm_rf(self, gfile):
         try:
-            to_delete = [f for f in self._recurse(gfile, dir_included=True)]
+            to_delete = [
+                f for f in self._recurse(
+                    gfile, dir_included=True, follow_symlinks=False
+                )
+            ]
             # make sure to delete the parent directory last:
             to_delete.sort(reverse=True, key=lambda f: f.get_uri())
             for f in to_delete:
@@ -444,31 +443,26 @@ class Plugin(openpaperwork_core.fs.CommonFsPluginBase):
             raise IOError(str(exc))
 
     def fs_get_mtime(self, url):
-        if not self._is_file_uri(url):
-            return None
-
         try:
             f = self.vfs.get_file_for_uri(url)
             if not f.query_exists():
                 raise IOError("File {} does not exist".format(str(url)))
+            if os.name == "nt":
+                # WORKAROUND(Jflesch):
+                # On Windows+MSYS2, it seems Gio.File.query_info()
+                # return always 0 for Gio.FILE_ATTRIBUTE_TIME_CHANGED.
+                path = self.fs_unsafe(url)
+                return int(os.stat(path).st_mtime)
             fi = f.query_info(
                 Gio.FILE_ATTRIBUTE_TIME_CHANGED, Gio.FileQueryInfoFlags.NONE
             )
             r = fi.get_attribute_uint64(Gio.FILE_ATTRIBUTE_TIME_CHANGED)
             if int(r) != 0:
                 return r
-            # WORKAROUND(Jflesch):
-            # On Windows+MSYS2, it seems Gio.File.query_info()
-            # return always 0 for Gio.FILE_ATTRIBUTE_TIME_CHANGED.
-            path = self.fs_unsafe(url)
-            return os.stat(path).st_mtime
         except GLib.GError as exc:
             LOGGER.warning("Gio.Gerror", exc_info=exc)
 
     def fs_getsize(self, url):
-        if not self._is_file_uri(url):
-            return None
-
         try:
             f = self.vfs.get_file_for_uri(url)
             fi = f.query_info(
@@ -480,7 +474,7 @@ class Plugin(openpaperwork_core.fs.CommonFsPluginBase):
             raise IOError(str(exc))
 
     def fs_isdir(self, url):
-        if not self._is_file_uri(url):
+        if not GLIB_AVAILABLE:
             return None
 
         try:
@@ -498,14 +492,6 @@ class Plugin(openpaperwork_core.fs.CommonFsPluginBase):
             raise IOError(str(exc))
 
     def fs_copy(self, old_url, new_url):
-        old_type = old_url.split(":", 1)[0]
-        new_type = new_url.split(":", 1)[0]
-        if old_type != new_type:
-            # use the more generic and cross-FS method
-            return super().fs_copy(old_url, new_url)
-
-        if not self._is_file_uri(old_url) or not self._is_file_uri(new_url):
-            return None
         try:
             old = self.vfs.get_file_for_uri(old_url)
             new = self.vfs.get_file_for_uri(new_url)
@@ -518,7 +504,7 @@ class Plugin(openpaperwork_core.fs.CommonFsPluginBase):
             return None
 
     def fs_mkdir_p(self, url):
-        if not self._is_file_uri(url):
+        if not GLIB_AVAILABLE:
             return None
 
         try:
@@ -538,13 +524,16 @@ class Plugin(openpaperwork_core.fs.CommonFsPluginBase):
 
         return True
 
-    def _recurse(self, parent, dir_included=False):
+    def _recurse(self, parent, dir_included=False, follow_symlinks=True):
         """
         Yield all the children (depth first), but not the parent.
         """
         try:
             children = parent.enumerate_children(
-                Gio.FILE_ATTRIBUTE_STANDARD_NAME,
+                (
+                    Gio.FILE_ATTRIBUTE_STANDARD_NAME
+                    + "," + Gio.FILE_ATTRIBUTE_STANDARD_IS_SYMLINK
+                ),
                 Gio.FileQueryInfoFlags.NOFOLLOW_SYMLINKS,
                 None
             )
@@ -555,9 +544,17 @@ class Plugin(openpaperwork_core.fs.CommonFsPluginBase):
 
         for child in children:
             name = child.get_name()
+            if not follow_symlinks:
+                is_symlink = child.has_attribute("standard::is-symlink")
+                is_symlink = is_symlink and child.get_is_symlink()
+                if is_symlink:
+                    yield parent.get_child(name)
+                    continue
             child = parent.get_child(name)
             try:
-                for sub in self._recurse(child, dir_included=dir_included):
+                for sub in self._recurse(
+                        child, dir_included=dir_included,
+                        follow_symlinks=follow_symlinks):
                     yield sub
             except GLib.GError:
                 yield child
@@ -566,20 +563,14 @@ class Plugin(openpaperwork_core.fs.CommonFsPluginBase):
             yield parent
 
     def fs_recurse(self, parent_uri, dir_included=False):
-        if not self._is_file_uri(parent_uri):
-            return None
-
         parent = self.vfs.get_file_for_uri(parent_uri)
         for f in self._recurse(parent, dir_included):
             yield f.get_uri()
 
     def fs_hide(self, uri):
-        if not self._is_file_uri(uri):
-            return None
-
         if os.name != 'nt':
             LOGGER.warning("fs_hide('%s') can only works on Windows", uri)
-            return
+            return None
         filepath = self.fs_unsafe(uri)
         LOGGER.info("Hiding file: {}".format(filepath))
         ret = ctypes.windll.kernel32.SetFileAttributesW(
@@ -587,11 +578,9 @@ class Plugin(openpaperwork_core.fs.CommonFsPluginBase):
         )
         if not ret:
             raise ctypes.WinError()
+        return True
 
     def fs_get_mime(self, uri):
-        if not self._is_file_uri(uri):
-            return None
-
         if os.name == 'nt':
             # WORKAROUND(Jflesch):
             # Gio.File.query_info().get_content_type() returns crap on Windows
@@ -605,16 +594,19 @@ class Plugin(openpaperwork_core.fs.CommonFsPluginBase):
         return info.get_content_type()
 
     def fs_mktemp(self, prefix=None, suffix=None, mode='w+b', **kwargs):
-        tmp = tempfile.NamedTemporaryFile(
-            prefix=prefix, suffix=suffix, delete=False, mode=mode
-        )
+        if 'b' not in mode:
+            tmp = tempfile.NamedTemporaryFile(
+                prefix=prefix, suffix=suffix, delete=False, mode=mode,
+                encoding='utf-8'
+            )
+        else:
+            tmp = tempfile.NamedTemporaryFile(
+                prefix=prefix, suffix=suffix, delete=False, mode=mode
+            )
         self.tmp_files.add(tmp.name)
         return (self.fs_safe(tmp.name), tmp)
 
     def fs_iswritable(self, url):
-        if not self._is_file_uri(url):
-            return None
-
         try:
             f = self.vfs.get_file_for_uri(url)
             fi = f.query_info(
